#include "NFC.h"
#include <SPI.h>
#include <PubSubClient.h>
#include "helpers.h"
#include <ArduinoLog.h>
#include <Adafruit_PN532.h>

NFC::NFC(int pin_sda, int pin_scl)
{
    Log.trace(F("NFC: Start"));

    // Create a new PN532 instance with the defined IRQ and RESET pins.
    rfid = new Adafruit_PN532(PN532_IRQ, PN532_RESET);
    rfid->begin(); // Initialize the PN532 chip

    // Retrieve the firmware version from the PN532.
    uint32_t versiondata = rfid->getFirmwareVersion();
    if (!versiondata)
    {
        // If no version data is returned, the PN532 was not detected.
        Log.error(F("Didn't find PN53x board"));
        while (true)
        {
            yield(); // Halt execution indefinitely if the board is missing.
        }
    }

    // Print chip and firmware details using formatted output.
    Log.trace(F("NFC: Found chip PN5%02X"), (versiondata >> 24) & 0xFF);
    Log.trace(F("NFC: Firmware ver. %d.%d"), (versiondata >> 16) & 0xFF, (versiondata >> 8) & 0xFF);
}

// Convert the card's UID (stored in a byte array) into a hexadecimal String.
String NFC::getUID() const
{
    String uidStr;
    uidStr.reserve(uidLength * 2); // Reserve memory to optimize String allocation.

    for (uint8_t i = 0; i < uidLength; i++)
    {
        // Ensure each byte is represented by two hex digits.
        if (uid[i] < 0x10)
        {
            uidStr += "0";
        }
        uidStr += String(uid[i], HEX); // Convert byte to hex string.
    }
    return uidStr;
}

// Returns true if a card has been successfully selected/connected.
bool NFC::hasCardSelected() const
{
    return cardSelected;
}

// This function is a placeholder for a possible PN532 self-test.
// Uncomment and implement if your PN532 supports self-testing.
bool NFC::testNFC()
{
    // return rfid->PCD_PerformSelfTest();
    return true;
}

// Check for the presence of an NFC card within a 1000ms timeout.
bool NFC::checkforCard()
{
    // Attempt to read the card's UID (ISO14443A standard) within 1000 ms.
    if (!rfid->readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength, 1000))
    {
        Log.warning("No card found within the timeout period");
        return false;
    }

    // Validate the card's UID length: 7 for DESFire, 4 for MIFARE Classic/ISO14443A.
    if (uidLength != 7 && uidLength != 4)
    {
        Log.warning("No card detected");
        return false;
    }

    // Report the type of card detected.
    if (uidLength == 7)
    {
        Log.info("DESFire card detected");
    }
    else
    {
        Log.info("MIFARE Classic or another ISO14443A card detected");
    }

    // Print the card's UID in hexadecimal format.
    Log.trace(F("UID: %s"), getUID().c_str());
    return true;
}

// Connect to the detected card using RATS (Request for ATS) and PPS (Protocol and Parameter Selection).
bool NFC::connecttoCard()
{
    // Send the RATS command via inDataExchange.
    // inDataExchange() returns a negative value on error.

    // --- RATS (Request for ATS) ---
    // RATS command structure: [0xE0, parameter]
    // The parameter byte encodes the Frame Size and other options.
    // 0xE0 identifies the command; 0x80 is an example parameter for frame size.
    uint8_t ratsCmd[] = {0xE0, 0x80};
    uint8_t atsResponse[16] = {0}; // Buffer to store ATS response.
    uint8_t atsResponseLength = sizeof(atsResponse);

    // Send the RATS command and check for errors.
    if (rfid->inDataExchange(ratsCmd, sizeof(ratsCmd), atsResponse, &atsResponseLength) < 0)
    {
        Log.error(F("Failed ATS"));
        // PN532 does not offer a dedicated halt command,
        // so you may simply return false to abort communication.
        return false;
    }

    // (Optional) Process the ATS response here if needed.
    // The ATS provides card-specific information, such as supported protocols and frame sizes.

    // --- PPS (Protocol and Parameter Selection) ---
    // PPS command example: [0xD0, PPS0, PPS1]
    // 0xD0 indicates the PPS command; 0x11 and 0x00 are example parameters.
    uint8_t ppsCmd[] = {0xD0, 0x11, 0x00};
    uint8_t ppsResponse[16] = {0}; // Buffer to store PPS response.
    uint8_t ppsResponseLength = sizeof(ppsResponse);

    // Send the PPS command and check for errors.
    if (rfid->inDataExchange(ppsCmd, sizeof(ppsCmd), ppsResponse, &ppsResponseLength) < 0)
    {
        Log.error(F("Failed PPS"));
        return false;
    }

    // If both RATS and PPS commands succeed, mark the card as selected.
    cardSelected = true;

    // Save any card-specific data if needed here.
    // Note: Unlike MFRC522, Adafruit_PN532 does not store the UID in a public member.
    // We save the UID when we detect the card.
    pcb = 0x0A; // Example PCB (Protocol Control Byte) value.
    return true;
}

// Disconnect from the card by issuing an ISO14443A HALT command.
bool NFC::disconnectCard()
{
    // HALT command for ISO14443A cards in auto-CRC mode: [0x50, 0x00]
    // You can either compute the CRC or, if your setup automatically appends it,
    // simply send the command without CRC bytes.
    // Here, we assume you have to include the CRC.
    // Note: The correct CRC for [0x50, 0x00] is usually 0x00, 0x00 if the PN532
    // is set to auto-calculate CRC, or you can compute it if needed.

    // Option 1: If your PN532 is configured to append the CRC automatically,
    // you can send just the two-byte command.
    uint8_t haltCmd[] = {0x50, 0x00};
    uint8_t response[8] = {0}; // Buffer to hold any response.
    uint8_t responseLength = sizeof(response);

    // Send the HALT command. A negative return value indicates failure.
    int status = rfid->inDataExchange(haltCmd, sizeof(haltCmd), response, &responseLength);
    if (status < 0)
    {
        Log.error(F("Failed to send HALT command: %d"), status); // Combined Serial.println calls
        return false;
    }

    // Option 2: If you need to send a complete command with CRC, you must
    // calculate the CRC bytes for [0x50, 0x00]. For many ISO14443A cards, the CRC
    // may be 0x00, 0x00 when using a PN532 in auto-CRC mode, so this might work:
    //
    // uint8_t haltCmd[] = { 0x50, 0x00, 0x00, 0x00 };
    // int status = rfid->inDataExchange(haltCmd, sizeof(haltCmd), response, &responseLength);
    // if (status < 0) { ... }
    //
    // Uncomment the above block and comment out the Option 1 block if needed.

    return true;
}

// A placeholder function to test card functionalities; extend as needed.
bool NFC::testCard()
{
    return true;
}

uint8_t NFC::Transceive(uint8_t *command, uint8_t command_len, uint8_t *response, uint8_t *response_len)
{
    // Check if command length exceeds the maximum allowed (command_len + 2 <= 255)
    if (command_len > 253)
    {
        return 0x01; // General error (no room for the command)
    }

    // Construct the request buffer with PCB and CID
    uint8_t request_buffer[command_len + 2]; // PCB + CID + command data
    request_buffer[0] = pcb;                 // Protocol Control Byte
    request_buffer[1] = cid;                 // Card Identifier
    memcpy(&request_buffer[2], command, command_len);

    // Toggle PCB as per ISO-DEP protocol
    pcb = (pcb == 0x0A) ? 0x0B : 0x0A;

    // Prepare response buffer
    uint8_t response_buffer[128]; // Adjust buffer size as needed
    // Clamp the size to 255 because inDataExchange expects a uint8_t
    uint8_t response_length = (sizeof(response_buffer) > 255) ? 255 : sizeof(response_buffer);

    // Transceive data using Adafruit's inDataExchange
    uint8_t status = rfid->inDataExchange(request_buffer, command_len + 2, response_buffer, &response_length);

    // Check if the operation was successful
    if (status != 0x00)
    {                  // 0x00 is PN532_STATUS_OK
        return status; // Return the Adafruit_PN532 error code
    }

    // Check if response contains at least PCB and CID
    if (response_length < 2)
    {
        return 0x01; // Invalid response
    }

    // Extract the data part (excluding PCB and CID)
    *response_len = response_length - 2;
    memcpy(response, response_buffer + 2, *response_len);

    return 0x00; // Success (PN532_STATUS_OK)
}